@@ -133,7 +133,8 @@ def generate_model_mapping(model, input_is_sparse):
                                    f" {layer_neural.name}. Only one pooling "
                                    f"layer is supported.")
             layer_ak.index_pool = i
-        elif isinstance(layer, layers.BatchNormalization):
+        elif (isinstance(layer, layers.BatchNormalization) or
+              layer.__class__.__name__ == "BatchNormalization"):
             if layer_ak.index_batchnorm:
                 raise RuntimeError(f"Two BatchNormalization layers were "
                                    f"detected after layer {layer_neural.name}."
@@ -168,8 +169,12 @@ def generate_model_mapping(model, input_is_sparse):
             continue
         else:
             # If you got here it means the layer is not recognised: raise an error.
-            raise RuntimeError(f"layer {layer.name}: unsupported type "
-                               f"{layer.__class__.__name__}.")
+            message = ""
+            if layer.__class__.__name__ == 'ReLU':
+                message = ("\nYou must use a quantized activation layer, like "
+                           "ActivationDiscreteRelu, instead of ReLU.")
+            raise RuntimeError(f"Layer {layer.name}: unsupported type "
+                               f"{layer.__class__.__name__}." + message)
 
     # Append last parsed layer if any
     layer_maps.append(layer_ak)
@@ -217,10 +222,10 @@ def check_mapping_compatibility(model_map):
     for layer_map in model_map.layer_maps:
         layer_neural = layers_k[layer_map.index_neural]
 
-        # Raise error if BatchNormalization 'axis' is different from the last
-        # dimension. The 'axis' parameter is a list containing the axes on what
-        # the batch normalization is applied.
         if layer_map.index_batchnorm:
+            # Raise error if BatchNormalization 'axis' is different from the last
+            # dimension. The 'axis' parameter is a list containing the axes on what
+            # the batch normalization is applied.
             layer_BN = layers_k[layer_map.index_batchnorm]
             axis = layer_BN.axis
             if (len(layer_BN.axis) != 1 or
@@ -228,6 +233,22 @@ def check_mapping_compatibility(model_map):
                 raise RuntimeError(f"The BatchNormalization layer "
                                    f"{layer_BN.name} must be applied on the "
                                    f"last axis. Receives {layer_BN.axis}.")
+            # Raise error if a gamma is zero.
+            gammas = layer_BN.get_weights()[0]
+            if (gammas == 0).any():
+                raise RuntimeError(
+                    f"The BatchNormalization layer {layer_BN.name} has at least"
+                    f" one gamma equal to zero. This case is not supported.")
+            # Raise error if BatchNormalization has at least one negative gamma
+            # and if a MaxPool2D layer is placed before it.
+            if (gammas <= 0).any() and layer_map.index_pool:
+                layer_pool = layers_k[layer_map.index_pool]
+                if (isinstance(layer_pool, layers.MaxPool2D) and
+                    (layer_map.index_pool < layer_map.index_batchnorm)):
+                    raise RuntimeError(
+                        f"The BatchNormalization layer {layer_BN.name} has at "
+                        f"least one negative gamma and a MaxPool2D layer is "
+                        f"placed before it. This case is not supported.")
 
         # Raise error if BatchNormalization is placed after the activation
         if (layer_map.index_batchnorm and layer_map.index_activation and
@@ -409,9 +430,6 @@ def check_model_compatibility(model_keras, input_is_sparse):
     - all layers with the 'data_format' parameter must be 'channels_last'
     - all processing quantized layers and ActivationDiscreteRelu must have a
       valid quantization bitwidth
-    - a QuantizedConv2D layer must have strides of (1, 1). Exception: if the
-      inputs are images, the first convolutional layer can have any value
-      for strides.
     - a QuantizedDense layer must have an input shape of (N,) or (1, 1, N)
     - a BatchNormalization layer must have 'axis' set to -1 (default)
     - a BatchNormalization layer cannot have negative gammas
