@@ -35,7 +35,7 @@ def _get_convolution_mode(str_conv_mode):
 def _parse_input_data(layer, params):
     params["input_height"] = int(layer.input_shape[1])
     params["input_width"] = int(layer.input_shape[2])
-    params["input_features"] = int(layer.input_shape[3])
+    params["input_channels"] = int(layer.input_shape[3])
     params['name'] = layer.name + "_input"
 
 
@@ -62,14 +62,13 @@ def _parse_conv(layer, params):
     if not isinstance(layer, qlayers.QuantizedConv2D):
         raise TypeError(f"Layer {layer.name} was expected to be "
                         "QuantizedConv2D")
-    if layer.strides != (1, 1):
-        raise ValueError(f"Layer {layer.name} must have strides of (1,1). "
-                         f"Received strides of {layer.strides}.")
     params["convolution_mode"] = _get_convolution_mode(layer.padding)
     params["kernel_height"] = layer.kernel_size[0]
     params["kernel_width"] = layer.kernel_size[1]
     params["num_neurons"] = int(layer.kernel.shape[3])
     params["weights_bits"] = layer.quantizer.bitwidth
+    params["stride_x"] = layer.strides[1]
+    params["stride_y"] = layer.strides[0]
     params["name"] = layer.name
 
 
@@ -77,13 +76,17 @@ def _parse_separable_conv(layer, params):
     if not isinstance(layer, qlayers.QuantizedSeparableConv2D):
         raise TypeError(f"Layer {layer.name} was expected to be "
                         "QuantizedSeparableConv2D")
+    if layer.quantizer_dw.bitwidth != layer.quantizer.bitwidth:
+        raise ValueError(f"Quantized layer {layer.name} must have the same "
+                         f"bitwidth for depthwise and pointwise quantizers.")
     params["convolution_mode"] = _get_convolution_mode(layer.padding)
     params["kernel_height"] = layer.kernel_size[0]
     params["kernel_width"] = layer.kernel_size[1]
     # num neurons is set to the number of filters of the depthwise
-    params["num_neurons"] = int(layer.depthwise_kernel.shape[2])
-    params["num_pointwise_neurons"] = int(layer.pointwise_kernel.shape[3])
+    params["num_neurons"] = int(layer.pointwise_kernel.shape[3])
     params["weights_bits"] = layer.quantizer.bitwidth
+    params["stride_x"] = layer.strides[1]
+    params["stride_y"] = layer.strides[0]
     params["name"] = layer.name
 
 
