@@ -16,14 +16,17 @@
 # ******************************************************************************
 
 # Tensorflow imports
+import tensorflow as tf
 from tensorflow.python.keras import backend as K
 from tensorflow.keras import layers
+from tensorflow.keras.utils import serialize_keras_object
 from tensorflow.python.keras.utils import conv_utils
 from tensorflow.python.ops import math_ops
 from tensorflow.python.ops import nn
 from tensorflow.python.ops import sparse_ops
 from tensorflow.python.ops import standard_ops, gen_math_ops
 from tensorflow.python.eager import context
+from .quantization_ops import get as get_quantizer
 from .quantization_ops import WeightFloat, BaseWeightQuantizer, ceil_through
 
 
@@ -50,6 +53,9 @@ class QuantizedConv2D(layers.Conv2D):
     the forward pass.
 
     """
+    unsupported_args = {
+        'data_format', 'activation', 'dilation_rate', 'activity_regularizer'
+    }
 
     def __init__(self,
                  filters,
@@ -90,13 +96,8 @@ class QuantizedConv2D(layers.Conv2D):
                 to apply during the forward pass.
 
         """
-        unsupported_args = {
-            'data_format', 'activation', 'dilation_rate', 'activity_regularizer'
-        }
-        _check_unsupported_args(kwargs, unsupported_args)
-        if not isinstance(quantizer, BaseWeightQuantizer):
-            raise AttributeError("Quantizer object should be used")
-        self.quantizer = quantizer
+        _check_unsupported_args(kwargs, self.unsupported_args)
+        self.quantizer = get_quantizer(quantizer)
         super(QuantizedConv2D,
               self).__init__(filters=filters,
                              kernel_size=kernel_size,
@@ -131,6 +132,13 @@ class QuantizedConv2D(layers.Conv2D):
 
         return outputs
 
+    def get_config(self):
+        config = super().get_config()
+        config['quantizer'] = serialize_keras_object(self.quantizer)
+        for kwarg in self.unsupported_args:
+            config.pop(kwarg, None)
+        return config
+
 
 class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
     """A quantization-aware Keras depthwise convolutional layer.
@@ -139,9 +147,13 @@ class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
     weights during the forward pass.
 
     """
+    unsupported_args = {
+        'data_format', 'activation', 'depth_multiplier', 'activity_regularizer'
+    }
 
     def __init__(self,
                  kernel_size,
+                 strides=(1, 1),
                  padding='valid',
                  use_bias=True,
                  depthwise_initializer='glorot_uniform',
@@ -156,6 +168,8 @@ class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
 
         Args:
             kernel_size (a tuple of integer): the kernel spatial dimensions.
+            strides (integer, or tuple of integers, optional): strides of the
+                convolution along height and width.
             padding (str, optional): One of 'valid' or 'same'.
             use_bias (boolean, optional): whether the layer uses a bias vector.
             depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
@@ -174,16 +188,11 @@ class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
                 to apply during the forward pass.
 
         """
-        unsupported_args = {
-            'data_format', 'activation', 'strides', 'depth_multiplier',
-            'activity_regularizer'
-        }
-        _check_unsupported_args(kwargs, unsupported_args)
-        if not isinstance(quantizer, BaseWeightQuantizer):
-            raise AttributeError("Quantizer object should be used")
-        self.quantizer = quantizer
+        _check_unsupported_args(kwargs, self.unsupported_args)
+        self.quantizer = get_quantizer(quantizer)
         super(QuantizedDepthwiseConv2D,
               self).__init__(kernel_size=kernel_size,
+                             strides=strides,
                              padding=padding,
                              use_bias=use_bias,
                              depthwise_initializer=depthwise_initializer,
@@ -216,6 +225,13 @@ class QuantizedDepthwiseConv2D(layers.DepthwiseConv2D):
                                   dilation_rate=self.dilation_rate,
                                   data_format=self.data_format)
 
+    def get_config(self):
+        config = super().get_config()
+        config['quantizer'] = serialize_keras_object(self.quantizer)
+        for kwarg in self.unsupported_args:
+            config.pop(kwarg, None)
+        return config
+
 
 class QuantizedDense(layers.Dense):
     """A quantization-aware Keras dense layer.
@@ -224,6 +240,7 @@ class QuantizedDense(layers.Dense):
     the forward pass.
 
     """
+    unsupported_args = {'activation', 'activity_regularizer'}
 
     def __init__(self,
                  units,
@@ -257,11 +274,8 @@ class QuantizedDense(layers.Dense):
                 to apply during the forward pass.
 
         """
-        unsupported_args = {'activation', 'activity_regularizer'}
-        _check_unsupported_args(kwargs, unsupported_args)
-        if not isinstance(quantizer, BaseWeightQuantizer):
-            raise AttributeError("Quantizer object should be used")
-        self.quantizer = quantizer
+        _check_unsupported_args(kwargs, self.unsupported_args)
+        self.quantizer = get_quantizer(quantizer)
         super(QuantizedDense,
               self).__init__(units=units,
                              use_bias=use_bias,
@@ -306,6 +320,13 @@ class QuantizedDense(layers.Dense):
             outputs = nn.bias_add(outputs, self.bias)
         return outputs
 
+    def get_config(self):
+        config = super().get_config()
+        config['quantizer'] = serialize_keras_object(self.quantizer)
+        for kwarg in self.unsupported_args:
+            config.pop(kwarg, None)
+        return config
+
 
 class QuantizedSeparableConv2D(layers.SeparableConv2D):
     """A quantization-aware Keras separable convolutional layer.
@@ -314,10 +335,15 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
     weights during the forward pass.
 
     """
+    unsupported_args = {
+        'data_format', 'activation', 'dilation_rate', 'depth_multiplier',
+        'activity_regularizer'
+    }
 
     def __init__(self,
                  filters,
                  kernel_size,
+                 strides=(1, 1),
                  padding='valid',
                  use_bias=True,
                  depthwise_initializer='glorot_uniform',
@@ -330,12 +356,15 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
                  pointwise_constraint=None,
                  bias_constraint=None,
                  quantizer=WeightFloat(),
+                 quantizer_dw=None,
                  **kwargs):
         """Creates a quantization-aware separable convolutional layer.
 
         Args:
             filters (integer): the number of filters.
             kernel_size (tuple of integer): the kernel spatial dimensions.
+            strides (integer, or tuple of integers, optional): strides of the
+                convolution along height and width.
             padding (str, optional): One of 'valid' or 'same'.
             use_bias (boolean, optional): Whether the layer uses a bias vector.
             depthwise_initializer (str, or a :obj:`tf.keras.initializer`, optional):
@@ -360,17 +389,20 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
                 during the forward pass.
 
         """
-        unsupported_args = {
-            'data_format', 'activation', 'strides', 'dilation_rate',
-            'depth_multiplier', 'activity_regularizer'
-        }
-        _check_unsupported_args(kwargs, unsupported_args)
-        if not isinstance(quantizer, BaseWeightQuantizer):
-            raise AttributeError("Quantizer object should be used")
-        self.quantizer = quantizer
+        _check_unsupported_args(kwargs, self.unsupported_args)
+        self.quantizer = get_quantizer(quantizer)
+        if quantizer_dw is None:
+            # If no depthwise quantizer provided, use the pointwise quantizer
+            # Note: this is compatible with legacy models
+            self.quantizer_dw = get_quantizer(
+                serialize_keras_object(self.quantizer))
+        else:
+            self.quantizer_dw = get_quantizer(quantizer_dw)
+
         super(QuantizedSeparableConv2D,
               self).__init__(filters=filters,
                              kernel_size=kernel_size,
+                             strides=strides,
                              padding=padding,
                              use_bias=use_bias,
                              depthwise_initializer=depthwise_initializer,
@@ -400,7 +432,7 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
         strides = (1,) + self.strides + (1,)
         outputs = nn.separable_conv2d(
             inputs,
-            self.quantizer.quantize(self.depthwise_kernel),
+            self.quantizer_dw.quantize(self.depthwise_kernel),
             self.quantizer.quantize(self.pointwise_kernel),
             strides=strides,
             padding=self.padding.upper(),
@@ -416,6 +448,14 @@ class QuantizedSeparableConv2D(layers.SeparableConv2D):
 
         return outputs
 
+    def get_config(self):
+        config = super().get_config()
+        config['quantizer_dw'] = serialize_keras_object(self.quantizer_dw)
+        config['quantizer'] = serialize_keras_object(self.quantizer)
+        for kwarg in self.unsupported_args:
+            config.pop(kwarg, None)
+        return config
+
 
 class BaseQuantizedActivation(layers.Activation):
     """Base class for quantized activation layers."""
@@ -431,6 +471,7 @@ class ActivationDiscreteRelu(BaseQuantizedActivation):
     [0,6].
 
     """
+    unsupported_args = {'activation'}
 
     def __init__(self, bitwidth=1, **kwargs):
         """Creates a discrete ReLU for the specified bitwidth.
@@ -439,6 +480,8 @@ class ActivationDiscreteRelu(BaseQuantizedActivation):
             bitwidth (int): the activation bitwidth.
 
         """
+        _check_unsupported_args(kwargs, self.unsupported_args)
+
         if bitwidth <= 0:
             raise ValueError("Activation 'bitwidth' must be greater than zero."
                              f"Receives 'bitwidth' {bitwidth}.")
@@ -449,23 +492,121 @@ class ActivationDiscreteRelu(BaseQuantizedActivation):
         self.gamma_k = self.relumax_ / levels
         self.t0_k = 0.5 * self.relumax_ / levels
         self.step = (2.**bitwidth * self.gamma_k) / 16
-        # Legacy parameters, used by cnn2snn.keras2akida: they might be removed
-        # later
-        self.scale_factor = levels / self.relumax_
-        self.threshold = self.t0_k
 
-        super().__init__(activation=self.activation, **kwargs)
+        super().__init__(activation=self.quantized_activation, **kwargs)
 
-    def activation(self, x):
+    def quantized_activation(self, x):
         """Evaluates the activations for the specified input Tensor.
 
         Args:
             x (:obj:`tensorflow.Tensor`): the input values.
 
         """
-        ceiled_scaled = ceil_through(x / self.gamma_k - 0.5)
+        ceiled_scaled = ceil_through((x - self.t0_k) / self.gamma_k)
         return K.clip(self.gamma_k * ceiled_scaled, 0, self.relumax_)
 
+    def get_config(self):
+        config = super().get_config()
+        config.update({'bitwidth': self.bitwidth})
+        for kwarg in self.unsupported_args:
+            config.pop(kwarg, None)
+        return config
+
+
+class QuantizedReLU(BaseQuantizedActivation):
+    """A Trainable Quantized ReLU Keras Activation.
+
+    Activations will be clipped to a trainable range, and quantized to a number
+    of values defined by the bitwidth: N = (2^bitwidth - 1) values plus zero
+
+    More specifically, this class uses two trainable variables:
+
+    - t0_k represents the lower bound of the activation range,
+    - gamma_k represents the step between two quantized activation values.
+
+    The activation range is therefore [t0_k, tN_k], with:
+
+        tN_k = t0_k + N * gamma_k = (2^bitwidth - 1) * gamma_k
+
+    In other words:
+
+    - inputs below t0_k will result in no activation
+    - inputs between t0_k and t0_k + tN_k will be ceiled to the nearest
+      t0_k + n * gamma_k, and result in a activation of n * gamma_k
+    - inputs above t0_k + tN_k will result in a activation of N * gamma_k
+
+    """
+    unsupported_args = {'activation'}
+
+    def __init__(self, bitwidth=1, **kwargs):
+        """Creates a QuantizedReLU for the specified bitwidth.
+
+        Args:
+            bitwidth (int): the activation bitwidth.
+
+        """
+        super().__init__(activation=self.quantized_activation, **kwargs)
+
+        if bitwidth <= 0:
+            raise ValueError("Activation 'bitwidth' must be greater than zero."
+                             f"Receives 'bitwidth' {bitwidth}.")
+
+        self.bitwidth = bitwidth
+        # Number of non-zero quantized activations
+        self.levels = 2.**bitwidth - 1
+        # To be compatible with models trained with ActivationDiscreteRelu, we
+        # keep the same rule for the initial value of the activation range
+        # relumax and we deduce the initial values of the t0 and gamma from it
+        relumax = min(self.levels, 6)
+        gamma_k = relumax / self.levels
+        # Create lower bound variable and initialize it
+        t0_k = gamma_k / 2.
+        self.t0_k_ = tf.Variable(name=f"{self.name}/t0_k",
+                                 initial_value=t0_k,
+                                 dtype=tf.float32,
+                                 trainable=True)
+        # We will actually train the rescaled step, whose value is more
+        # consistent between different bitwidth values
+        step = (2.**self.bitwidth) * gamma_k / 16
+        self.step_ = tf.Variable(name=f"{self.name}/step",
+                                 initial_value=step,
+                                 dtype=tf.float32,
+                                 trainable=True)
+
+    @property
+    def gamma_k(self):
+        return 16 * self.step / 2.**self.bitwidth
+
+    @property
+    def t0_k(self):
+        return self.t0_k_.numpy()
+
+    @property
+    def step(self):
+        return self.step_.numpy()
+
+    def quantized_activation(self, x):
+        """Evaluates the activations for the specified input Tensor.
+
+        Args:
+            x (:obj:`tensorflow.Tensor`): the input values.
+
+        """
+        # Evaluate integer activation
+        gamma_k = 16 * self.step_ / 2.**self.bitwidth
+        act = ceil_through((x - self.t0_k_) / gamma_k)
+        # Clip to the maximum value for that bitwidth
+        clip_act = tf.clip_by_value(act, 0, self.levels)
+        # Rescale to obtain the float value
+        return gamma_k * clip_act
+
+    def get_config(self):
+        config = super().get_config()
+        config.update({'bitwidth': self.bitwidth})
+        for kwarg in self.unsupported_args:
+            config.pop(kwarg, None)
+        return config
+
 
 # A helper to instantiate a Conv2D layer to which a modifier is assigned
 def conv2d(filters, kernel_size, modifier=None, **kwargs):
@@ -511,7 +652,3 @@ def activationDiscreteRelu(bitwidth=1, **kwargs):
 
 def activationDiscreteReluBits(bitwidth=1):
     return lambda **kwargs: activationDiscreteRelu(bitwidth, **kwargs)
-
-
-def batchNormalization(*args, **kwargs):
-    return layers.BatchNormalization(*args, **kwargs)
